from helga.db import db
from random import random
import jaraco.modb, math

def encode(target):
    """ Encode as mongodb friendly object """
    return jaraco.modb.encode(target)

def decode(target):
    """ Decode from mongodb """
    return jaraco.modb.decode(target)

def random_cursor(cursor):
    """ Pull a random item from a cursor """
    index = int(math.floor(random()*cursor.count()))
    return cursor.limit(1).skip(index).next()

def current_state():
    """ Return tuple representing hero, enemy, and in_encounter """
    hero = decode(db.quest.heroes.find_one())
    enemy = None
    in_encounter = db.quest.encounter.count() > 0
    if in_encounter:
        enemy = decode(db.quest.encounter.find_one())
    return (hero, enemy, in_encounter)



from helga.plugins import command
from helga_quest import driver

_help_text = """Collaboratively play an RPG from user driven content \
Usage: !quest (action|adventure|attack|mob|rest)\
!quest mob add '{"name":"Assault Shaker", "hp":1, "level":1, "xp":60}'\
!quest action add '{"name":"Assault Shaker", "description":"{name} peppers {target} for {dmg} damage", "attack":5}'\
!quest adventure\
helga> You've encountered a Assault Shaker!\
!quest attack\
helga> You strike for 1 damage, Assault Shaker peppers Hero for 5.9 damage\
!quest rest\
helga> You feel refreshed..."""


@command('quest', help=_help_text, shlex=True)
def quest(client, channel, nick, message, cmd, args):
    """ Parse commands and execute via driver """
    response = ''
    if len(args) == 0 or args[0] == 'status':
        response = driver.status()
    elif args[0] == 'action' or args[0] == 'actions':
        mod = args[2] if len(args) > 2 else ''
        response = driver.action(args[1], mod)
    elif args[0] == 'adventure':
        mob = args[1] if len(args) > 1 else ''
        response = driver.adventure(mob)
    elif args[0] == 'rest' or args[0] == 'sleep':
        response = driver.rest()
    elif args[0] == 'mob' or args[0] == 'mobs':
        response = driver.mob(args[1], args[2])
    elif args[0] == 'attack' or args[0] == 'fight':
        response = driver.attack()
    elif args[0] == 'drop':
        mod = args[1] if len(args) > 1 else ''
        response = driver.drop(mod)
    else:
        response = "I don't understand %s" % args[0]
    return response



import unittest
from helga_quest import core

class TestResults(unittest.TestCase):
    def setUp(self):
        self.roger = core.Being('Roger')
        self.roger.attack = 5
        self.enemy = core.Being('Enemy')
        self.enemy.xp = 2000

    def test_combat(self):
        """ verify combat """
        dmg = self.roger.do_attack(self.enemy)
        self.enemy.hp_current -= dmg
        self.assertTrue(dmg > 4)

    def test_defense(self):
        highdef = core.Being('Defense')
        highdef.defense = 200
        dmg = self.roger.do_attack(highdef)
        self.assertTrue(dmg < 2)

    def test_level(self):
        leveler = core.Being('Pwrlvl')
        leveler.xp = 5000
        while leveler.can_levelup():
            leveler.levelup()
        self.assertEquals(leveler.level, 7)

    def test_attack(self):
        action = core.Action('Atk', attack=5)
        dmg = self.roger.do_attack(self.enemy, attack_bonus=action.attack)
        self.assertTrue(dmg > 9)


if __name__ == '__main__':
    unittest.main()



from helga.db import db
from helga_quest.core import Being
from helga_quest.util import encode

VERSION = (1, 0, 0)

__version__ = '.'.join([str(v) for v in VERSION])

if db.quest.heroes.count() == 0:
    db.quest.heroes.insert(encode(Being(name='Hero', attack=5, hp=5, defense=20)))



""" Core classes to define rpg """
import math, random, re

class Action(object):
    """ A special attack a Being may execute against another Being """
    def __init__(self, name='Hero', description='{target} took {dmg} damage.',
                 hp=0, attack=0, defense=0, speed=0):
        self.name = name
        self.description = description
        self.hp = hp
        self.attack = attack
        self.defense = defense
        self.speed = speed

    def create_response(self, target, dmg):
        """ Create the string response the action should produce against the
        enemy. e.g. if the enemy's name is Phil, and his attack is Anthropy,
        the description might be
        'Phil's anthropy attack strikes to {target}'s heart, causing {dmg}
        damage'
        """
        kwargs = {}
        if re.search('{\s*dmg\s*}', self.description, re.I):
            kwargs['dmg']=dmg
        if re.search('{\s*target\s*}', self.description, re.I):
            kwargs['target']=target.name
        if re.search('{\s*name\s*}', self.description, re.I):
            kwargs['name']=self.name
        return self.description.format(**kwargs)

    def __unicode__(self):
        """ Unicode representation of action """
        return 'Action for {}, description: {}'.format(self.name, self.description)

    def __repr__(self):
        """ String representation of action """
        return self.__unicode__()


class Being(object):
    """ Something that lives and breathes and can die """
    def __init__(self, name='Hero', hp=5, attack=1, defense=1, speed=50,
                 level=1, xp=0):
        self.hp = hp
        self.attack = attack
        self.defense = defense
        self.speed = speed
        self.name = name
        self.level = level
        self.xp = xp
        self.hp_current = self.hp

    def do_attack(self, target, attack_bonus=0, defense_bonus=0):
        """ Weapon attack, returns dmg to be received. """
        return round((self.attack + attack_bonus) *
                     (100. / (100. + target.defense + defense_bonus)), 1)

    def can_levelup(self):
        """ Return true if being has enough experience to level up """
        return self.xp > Being.xp_to_level(self.level + 1)

    def levelup(self):
        """ Level up scale stats appropriately """
        self.level += 1
        self.hp += random.randint(5, 10)
        self.attack += random.randint(0, 3)
        self.defense += random.randint(0, 3)
        self.hp_current = self.hp

    def leveldown(self):
        """ Level down being """
        self.level -= 1
        self.hp -= random.randint(5, 10)
        self.attack -= random.randint(0, 3)
        self.defense -= random.randint(0, 3)
        self.hp_current = self.hp

    def scale_level(self, target_level):
        """ Scale difficulty to balance mob to target level """
        self.xp = self.xp / self.level * target_level
        self.xp = int(self.xp * (1 + random.gauss(0, .05)))
        while target_level > self.level:
            self.levelup()
        while target_level < self.level:
            self.leveldown()

    def initiative_roll(self, target_speed):
        """ Roll for initiative per round, returns true if won """
        self_roll = self.speed + random.gauss(50, 16)
        target_roll = target_speed + random.gauss(50, 16)
        return self_roll > target_roll

    def __unicode__(self):
        """ String representation of being """
        return '{} HP: {}/{} Level: {}'.format(self.name, self.hp_current, self.hp, self.level)

    def __repr__(self):
        """ String representation of being """
        return self.__unicode__()

    @staticmethod
    def xp_to_level(level):
        """ Calculate the amount of xp to get to the next level """
        if level <= 11:
            xp = 40 * math.pow(level, 2) + 360 * level
        elif level <= 27:
            xp = -.4 * math.pow(level, 3) + 40.4 * math.pow(level, 2) + 396 * level
        else:
            xp = (65 * math.pow(level, 2) - 165 * level - 6750) * .82
        return xp



from pip.req import parse_requirements
from setuptools import setup, find_packages

from pip.req import parse_requirements
from helga_quest import __version__ as version

requirements = [
    str(req.req) for req in parse_requirements('requirements.txt')
]

setup(
    name='helga-quest',
    version=version,
    description=('Collaborative RPG consisting of user driven content'),
    classifiers=[
        'Development Status :: 4 - Beta',
        'Topic :: Communications :: Chat :: Internet Relay Chat',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Communications :: Chat :: Internet Relay Chat'],
    keywords='irc bot quest',
    author='Jon Robison',
    author_email='narfman0@gmail.com',
    url='https://github.com/narfman0/helga-quest',
    license='LICENSE',
    packages=find_packages(),
    include_package_data=True,
    py_modules=['helga_quest.plugin'],
    zip_safe=True,
    install_requires=requirements,
    test_suite='tests.test_quest',
    entry_points = dict(
        helga_plugins=[
            'quest = helga_quest.plugin:quest',
        ],
    ),
)



""" Driver for core gameplay """
from helga.db import db
from helga_quest.core import Action, Being
from helga_quest.util import encode, decode, random_cursor, current_state
import json, random

def status():
    """ Return the current status of the game as string """
    hero, enemy, in_encounter = current_state()
    next_xp = Being.xp_to_level(hero.level + 1)
    response = '{} XP: {}/{}'.format(str(hero), hero.xp, next_xp)
    if not in_encounter:
        response += u", you are not actively fighting."
    else:
        response += ' vs %s.' % str(enemy)
    template = ' There are {} mobs and {} actions in database'
    response += template.format(db.quest.enemies.count(),
                                db.quest.actions.count())
    return response

def action(mod, statsstr):
    """ Add/remove actions for NPCs """
    stats = json.loads(statsstr)
    action = Action(**stats)
    if mod == 'add':
        db.quest.actions.insert(encode(action))
    elif mod == 'remove':
        db.quest.actions.remove({'name':action.name,
                                 'description':action.description})
    else:
        response = 'I do not understand %s' % mod
    return response

def adventure(mob=''):
    """ Begin new encounter """
    hero, enemy, in_encounter = current_state()
    if hero.hp_current <= 0:
        response = 'You must rest before adventuring!'
    elif in_encounter:
        response = "You can't abandon your valiant journey!"
    elif db.quest.enemies.count() == 0:
        response = "There are no enemies populated against which to quest!"
    else:
        if mob:
            query = db.quest.enemies.find({'name':mob})
        else:
            query = db.quest.enemies.find()
        enemy = decode(random_cursor(query))
        target_level = int(max(1, hero.level * (1 + random.gauss(0, .05))))
        enemy.scale_level(target_level)
        db.quest.encounter.insert(encode(enemy))
        template = "You've encountered a {} level {}!"
        response = template.format(enemy.name, enemy.level)
    return response

def rest():
    """ Direct player to rest if possible """
    hero, enemy, in_encounter = current_state()
    if in_encounter:
        response = "You can't rest whilst in combat!"
    else:
        hero.hp_current = hero.hp
        db.quest.heroes.remove({'name':hero.name})
        db.quest.heroes.insert(encode(hero))
        response = 'You feel refreshed and ready for combat'
    return response

def mob(mod, statsstr):
    """ Manage mobs in database """
    stats = json.loads(statsstr)
    being = Being(**stats)
    if mod == 'add':
        db.quest.enemies.insert(encode(being))
        response = 'Mob added'
    elif mod == 'remove':
        db.quest.enemies.remove({'name':being.name})
        response = 'Mob removed'
    return response

def execute_attack_enemy(hero, enemy, enemy_action):
    """ Execute enemy attack """
    received_dmg = enemy.do_attack(hero, attack_bonus=enemy_action.attack)
    hero.hp_current -= received_dmg
    db.quest.heroes.remove({'name':hero.name})
    db.quest.heroes.insert(encode(hero))
    response = enemy_action.create_response(hero, received_dmg) + ' '
    if hero.hp_current <= 0:
        db.quest.encounter.drop()
        response += '\nYou have been slain!'
    else:
        db.quest.encounter.remove({'name':enemy.name})
        db.quest.encounter.insert(encode(enemy))
    return response

def execute_attack_hero(hero, enemy, enemy_action):
    """ Execute hero attack """
    dmg = hero.do_attack(enemy, defense_bonus=enemy_action.defense)
    enemy.hp_current -= dmg
    db.quest.encounter.remove({'name':enemy.name})
    db.quest.encounter.insert(encode(enemy))
    response = 'You strike for %d damage. ' % dmg
    if enemy.hp_current <= 0:
        hero.xp += enemy.xp
        while hero.can_levelup():
            hero.levelup()
        db.quest.heroes.remove({'name':hero.name})
        db.quest.heroes.insert(encode(hero))
        db.quest.encounter.remove({'name':enemy.name})
        template = "You've slain the {}, and earned {} xp!"
        response += template.format(enemy.name, enemy.xp)
    return response

def attack():
    """ Execute attack on current encounter mob(s) """
    hero, enemy, in_encounter = current_state()
    if hero.hp_current <= 0:
        response = 'You must rest before exerting oneself!'
    elif not in_encounter:
        response = "There is no enemy to attack!"
    else:
        response = ''
        # grab enemy action to have defense bonus on hand
        query = db.quest.actions.find({'name':enemy.name})
        if query.count() > 0:
            action = decode(random_cursor(query))
        else:
            action = Action(name=enemy.name)
        if hero.initiative_roll(enemy.speed + action.speed):
            response += execute_attack_hero(hero, enemy, action)
            if enemy.hp_current > 0:
                response += execute_attack_enemy(hero, enemy, action)
        else:
            response += execute_attack_enemy(hero, enemy, action)
            if hero.hp_current > 0:
                response += execute_attack_hero(hero, enemy, action)

    return response

def drop(mod=''):
    """ Drop items from database """
    if not mod:
        db.quest.drop()
        db.quest.heroes.insert(encode(Being()))
        response = "Quest database dropped, so sad"
    elif mod == 'heroes':
        db.quest.heroes.drop()
        response = "Heroes dropped"
    elif mod == 'encounter':
        db.quest.encounter.drop()
        response = "Encounter dropped"
    elif mod == "enemies":
        db.quest.enemies.drop()
        response = "Enemies dropped"
    else:
        response = "I don't understand %s" % mod
    return response
